import is from '@sindresorhus/is';
import { RuntimeException } from '@poppinss/utils';
import { Route } from './route.js';
import { RouteGroup } from './group.js';
import { BriskRoute } from './brisk.js';
import { RoutesStore } from './store.js';
import { toRoutesJSON } from '../helpers.js';
import { RouteResource } from './resource.js';
import { LookupStore } from './lookup_store/main.js';
import { RouteMatchers as Matchers } from './matchers.js';
export class Router extends LookupStore {
    #app;
    #store = new RoutesStore();
    #globalMatchers = {};
    #middlewareStore;
    #openedGroups = [];
    routes = [];
    usingDomains = false;
    matchers = new Matchers();
    constructor(app, encryption, middlewareStore, qsParser) {
        super(encryption, qsParser);
        this.#app = app;
        this.#middlewareStore = middlewareStore;
    }
    #pushToRoutes(entity) {
        const openedGroup = this.#openedGroups[this.#openedGroups.length - 1];
        if (openedGroup) {
            openedGroup.routes.push(entity);
            return;
        }
        this.routes.push(entity);
    }
    route(pattern, methods, handler) {
        const route = new Route(this.#app, this.#middlewareStore, {
            pattern,
            methods,
            handler,
            globalMatchers: this.#globalMatchers,
        });
        this.#pushToRoutes(route);
        return route;
    }
    any(pattern, handler) {
        return this.route(pattern, ['HEAD', 'OPTIONS', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE'], handler);
    }
    get(pattern, handler) {
        return this.route(pattern, ['GET', 'HEAD'], handler);
    }
    post(pattern, handler) {
        return this.route(pattern, ['POST'], handler);
    }
    put(pattern, handler) {
        return this.route(pattern, ['PUT'], handler);
    }
    patch(pattern, handler) {
        return this.route(pattern, ['PATCH'], handler);
    }
    delete(pattern, handler) {
        return this.route(pattern, ['DELETE'], handler);
    }
    group(callback) {
        const group = new RouteGroup([], this.#middlewareStore);
        this.#pushToRoutes(group);
        this.#openedGroups.push(group);
        callback();
        this.#openedGroups.pop();
        return group;
    }
    resource(resource, controller) {
        const resourceInstance = new RouteResource(this.#app, this.#middlewareStore, {
            resource,
            controller,
            shallow: false,
            globalMatchers: this.#globalMatchers,
        });
        this.#pushToRoutes(resourceInstance);
        return resourceInstance;
    }
    shallowResource(resource, controller) {
        const resourceInstance = new RouteResource(this.#app, this.#middlewareStore, {
            resource,
            controller,
            shallow: true,
            globalMatchers: this.#globalMatchers,
        });
        this.#pushToRoutes(resourceInstance);
        return resourceInstance;
    }
    on(pattern) {
        const briskRoute = new BriskRoute(this.#app, this.#middlewareStore, {
            pattern,
            globalMatchers: this.#globalMatchers,
        });
        this.#pushToRoutes(briskRoute);
        return briskRoute;
    }
    where(param, matcher) {
        if (typeof matcher === 'string') {
            this.#globalMatchers[param] = { match: new RegExp(matcher) };
        }
        else if (is.regExp(matcher)) {
            this.#globalMatchers[param] = { match: matcher };
        }
        else {
            this.#globalMatchers[param] = matcher;
        }
        return this;
    }
    commit() {
        const routeNamesByDomain = new Map();
        toRoutesJSON(this.routes).forEach((route) => {
            if (!routeNamesByDomain.has(route.domain)) {
                routeNamesByDomain.set(route.domain, new Set());
            }
            const routeNames = routeNamesByDomain.get(route.domain);
            if (route.name && routeNames.has(route.name)) {
                throw new RuntimeException(`Route with duplicate name found. A route with name "${route.name}" already exists`);
            }
            if (route.name) {
                routeNames.add(route.name);
            }
            this.register(route);
            this.#store.add(route);
        });
        routeNamesByDomain.clear();
        this.usingDomains = this.#store.usingDomains;
        this.routes = [];
        this.#globalMatchers = {};
    }
    match(url, method, hostname) {
        const matchingDomain = this.#store.matchDomain(hostname);
        return matchingDomain.length
            ? this.#store.match(url, method, {
                tokens: matchingDomain,
                hostname: hostname,
            })
            : this.#store.match(url, method);
    }
    makeUrl(routeIdentifier, params, options) {
        const normalizedOptions = Object.assign({}, options);
        const builder = normalizedOptions.domain
            ? this.builderForDomain(normalizedOptions.domain)
            : this.builder();
        builder.params(params);
        builder.qs(normalizedOptions.qs);
        normalizedOptions.prefixUrl && builder.prefixUrl(normalizedOptions.prefixUrl);
        normalizedOptions.disableRouteLookup && builder.disableRouteLookup();
        return builder.make(routeIdentifier);
    }
    makeSignedUrl(routeIdentifier, params, options) {
        const normalizedOptions = Object.assign({}, options);
        const builder = normalizedOptions.domain
            ? this.builderForDomain(normalizedOptions.domain)
            : this.builder();
        builder.params(params);
        builder.qs(normalizedOptions.qs);
        normalizedOptions.prefixUrl && builder.prefixUrl(normalizedOptions.prefixUrl);
        normalizedOptions.disableRouteLookup && builder.disableRouteLookup();
        return builder.makeSigned(routeIdentifier, normalizedOptions);
    }
}
