import Middleware from '@poppinss/middleware';
import { RuntimeException } from '@poppinss/utils';
import { Qs } from '../qs.js';
import debug from '../debug.js';
import { Request } from '../request.js';
import { Response } from '../response.js';
import { Router } from '../router/main.js';
import { HttpContext } from '../http_context/main.js';
import { MiddlewareStore } from '../middleware/store.js';
import { finalHandler } from './factories/final_handler.js';
import { writeResponse } from './factories/write_response.js';
import { useReturnValue } from './factories/use_return_value.js';
import { asyncLocalStorage } from '../http_context/local_storage.js';
import { middlewareHandler } from './factories/middleware_handler.js';
export class Server {
    #errorHandler;
    #resolvedErrorHandler = {
        handle(error, ctx) {
            ctx.response.status(error.status || 500).send(error.message || 'Internal server error');
        },
    };
    #app;
    #encryption;
    #config;
    #qsParser;
    #serverMiddlewareStack;
    #router;
    #underlyingHttpServer;
    get usingAsyncLocalStorage() {
        return asyncLocalStorage.isEnabled;
    }
    constructor(app, encryption, config) {
        this.#app = app;
        this.#encryption = encryption;
        this.#config = config;
        this.#qsParser = new Qs(this.#config.qs);
        this.#createAsyncLocalStore();
        debug('server config: %O', this.#config);
    }
    #createAsyncLocalStore() {
        if (this.#config.useAsyncLocalStorage) {
            debug('creating http context async local storage');
            asyncLocalStorage.create();
        }
        else {
            asyncLocalStorage.destroy();
        }
    }
    #createRouter(middlewareStore) {
        this.#router = new Router(this.#app, this.#encryption, middlewareStore, this.#qsParser);
    }
    #createServerMiddlewareStack(middlewareStore) {
        this.#serverMiddlewareStack = new Middleware();
        middlewareStore.list().forEach((middleware) => this.#serverMiddlewareStack.add(middleware));
        this.#serverMiddlewareStack.freeze();
    }
    #createContext(req, res) {
        const request = new Request(req, res, this.#encryption, this.#config, this.#qsParser);
        const response = new Response(req, res, this.#encryption, this.#config, this.#router, this.#qsParser);
        return new HttpContext(request, response, this.#app.logger.child({}));
    }
    use(serverMiddleware, routerMiddleware, namedMiddleware) {
        this.#createRouter(new MiddlewareStore(routerMiddleware, namedMiddleware));
        this.#createServerMiddlewareStack(new MiddlewareStore(serverMiddleware));
        return this;
    }
    errorHandler(handler) {
        this.#errorHandler = handler;
        return this;
    }
    async boot() {
        debug('booting HTTP server');
        if (!this.#router) {
            throw new RuntimeException('Cannot boot HTTP server. Register middleware using "server.use" first');
        }
        this.#router.commit();
        if (this.#errorHandler) {
            if (debug.enabled) {
                debug('using custom error handler "%s"', this.#errorHandler);
            }
            const moduleExports = await this.#errorHandler();
            this.#resolvedErrorHandler = await this.#app.container.make(moduleExports.default);
        }
    }
    #handleRequest(ctx, resolver) {
        return this.#serverMiddlewareStack.runner()
            .finalHandler(finalHandler(this.#router, resolver, ctx))
            .run(middlewareHandler(resolver, ctx))
            .then(useReturnValue(ctx))
            .catch((error) => this.#resolvedErrorHandler.handle(error, ctx))
            .finally(writeResponse(ctx));
    }
    setNodeServer(server) {
        this.#underlyingHttpServer = server;
    }
    close() {
        return new Promise((resolve, reject) => {
            if (!this.#underlyingHttpServer || !this.#underlyingHttpServer.listening) {
                resolve();
                return;
            }
            this.#underlyingHttpServer.close((error) => {
                if (error) {
                    return reject(error);
                }
                resolve();
            });
        });
    }
    getNodeServer() {
        return this.#underlyingHttpServer;
    }
    getRouter() {
        return this.#router;
    }
    handle(req, res) {
        const ctx = this.#createContext(req, res);
        const resolver = this.#app.container.createResolver();
        if (this.usingAsyncLocalStorage) {
            return asyncLocalStorage.storage.run(ctx, () => this.#handleRequest(ctx, resolver));
        }
        return this.#handleRequest(ctx, resolver);
    }
}
